"""
PropertyField
=============
"""

import numpy as np
from ansys.dpf.core.common import natures, locations, _get_size_of_list
from ansys.dpf.core import scoping, dimensionality
from ansys.dpf.core.field_base import _FieldBase, _LocalFieldBase
from ansys.dpf.gate import (
    property_field_abstract_api,
    property_field_capi,
    property_field_grpcapi,
    dpf_array,
    dpf_vector,
)


class PropertyField(_FieldBase):
    """Describes field properties such as connectivity.

    This class is a field with integer values instead of double values.

    Parameters
    ----------
    nentities: int
        Number of entities that the property field is to contain. The
        default is ``0``.
    nature: core.natures
        Nature of the property field, such as scalar or vector.
    location : str, optional
        Location of the property field. Options are in :class:`ansys.dpf.core.locations`.
        The default is :class:`ansys.dpf.core.locations.nodal`.
    property field : Field, ansys.grpc.dpf.field_pb2.Field, ctypes.c_void_p, optional
        Field message generated from a gRPC stub, or returned by DPF's C clients.
    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Returns
    -------
    property_field: PropertyField

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> pfield = dpf.PropertyField()
    >>> list_ids = [1, 2, 4, 6, 7]
    >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
    >>> pfield.scoping = scop
    >>> list_data = [20, 30, 50, 70, 80]
    >>> pfield.data = list_data

    """

    def __init__(
        self,
        nentities=0,
        nature=natures.scalar,
        location=locations.nodal,
        property_field=None,
        server=None,
    ):
        super().__init__(
            nentities=nentities,
            nature=nature,
            location=location,
            field=property_field,
            server=server,
        )

    @property
    def _api(self) -> property_field_abstract_api.PropertyFieldAbstractAPI:
        if not self._api_instance:
            self._api_instance = self._server.get_api_for_type(
                capi=property_field_capi.PropertyFieldCAPI,
                grpcapi=property_field_grpcapi.PropertyFieldGRPCAPI,
            )
        return self._api_instance

    def _init_api_env(self):
        self._api.init_property_field_environment(self)

    @staticmethod
    def _field_create_internal_obj(
        api: property_field_abstract_api.PropertyFieldAbstractAPI,
        client,
        nature,
        nentities,
        location=locations.nodal,
        ncomp_n=0,
        ncomp_m=0,
        with_type=None,
    ):
        dim = dimensionality.Dimensionality([ncomp_n, ncomp_m], nature)
        if client is not None:
            return api.csproperty_field_new_on_client(
                client, nentities, nentities * dim.component_count
            )
        else:
            return api.csproperty_field_new(nentities, nentities * dim.component_count)

    @property
    def location(self):
        """Location of the property field.

        A property field contains a scoping, which is the location that is read.
        To update location, directly update the scoping location.

        Returns
        -------
        location : str
            Location string, can be found in :class:`ansys.dpf.core.locations`: ie.
            ``dpf.locations.nodal`` or ``dpf.locations.elemental``.

        Examples
        --------
        Create a property field and request the location.

        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.PropertyField()
        >>> list_ids = [1, 2, 4, 6, 7]
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.scoping.location = dpf.locations.nodal
        >>> pfield.location
        'Nodal'

        """
        if self.scoping:
            return self.scoping.location
        else:
            return None

    @location.setter
    def location(self, value):
        """Change the property field location.

        Parameters
        ----------
        location : str or locations
            Location string, can be found in :class:`ansys.dpf.core.locations`: ie.
            ``dpf.locations.nodal`` or ``dpf.locations.elemental``.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.PropertyField()
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.location = 'Nodal'
        >>> pfield.location
        'Nodal'

        """
        if self.scoping:
            self.scoping.location = value
        else:
            raise Exception(
                "Property field location is based on scoping, and scoping is not defined"
            )

    @property
    def component_count(self):
        return self._api.csproperty_field_elementary_data_size(self)

    @property
    def elementary_data_count(self):
        return self._api.csproperty_field_get_number_elementary_data(self)

    @property
    def size(self):
        return self._api.csproperty_field_get_data_size(self)

    def _set_scoping(self, scoping):
        self._api.csproperty_field_set_cscoping(self, scoping)

    def _get_scoping(self):
        return scoping.Scoping(
            scoping=self._api.csproperty_field_get_cscoping(self), server=self._server
        )

    def get_entity_data(self, index):
        try:
            vec = dpf_vector.DPFVectorInt(client=self._server.client)
            self._api.csproperty_field_get_entity_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, index
            )
            data = dpf_array.DPFArray(vec)

        except NotImplementedError:
            data = self._api.csproperty_field_get_entity_data(self, index)
        n_comp = self.component_count
        if n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def get_entity_data_by_id(self, id):
        try:
            vec = dpf_vector.DPFVectorInt(client=self._server.client)
            self._api.csproperty_field_get_entity_data_by_id_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size, id
            )
            data = dpf_array.DPFArray(vec)
        except NotImplementedError:
            index = self.scoping.index(id)
            if index < 0:
                raise ValueError(f"The ID {id} must be greater than 0.")
            data = self.get_entity_data(index)
        n_comp = self.component_count
        if n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def append(self, data, scopingid):
        self._api.csproperty_field_push_back(self, scopingid, _get_size_of_list(data), data)

    def _get_data_pointer(self):
        try:
            vec = dpf_vector.DPFVectorInt(client=self._server.client)
            self._api.csproperty_field_get_data_pointer_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            return dpf_array.DPFArray(vec)

        except NotImplementedError:
            return self._api.csproperty_field_get_data_pointer(self, True)

    def _set_data_pointer(self, data):
        return self._api.csproperty_field_set_data_pointer(self, _get_size_of_list(data), data)

    def _get_data(self, np_array=True):
        try:
            vec = dpf_vector.DPFVectorInt(client=self._server.client)
            self._api.csproperty_field_get_data_for_dpf_vector(
                self, vec, vec.internal_data, vec.internal_size
            )
            data = dpf_array.DPFArray(vec) if np_array else dpf_array.DPFArray(vec).tolist()
        except NotImplementedError:
            data = self._api.csproperty_field_get_data(self, np_array)
        n_comp = self.component_count
        if np_array and n_comp != 1 and data.size != 0:
            data.shape = (data.size // n_comp, n_comp)
        return data

    def _set_data(self, data):
        if isinstance(data, (np.ndarray, np.generic)):
            if (
                0 != self.size
                and self.component_count > 1
                and data.size // self.component_count != data.size / self.component_count
            ):
                raise ValueError(
                    f"An array of shape {self.shape} is expected and "
                    f"shape {data.shape} was input"
                )

            if data.dtype != np.int32:
                copy = np.empty_like(data, shape=data.shape, dtype=np.int32)
                copy[:] = data
                data = copy
        return self._api.csproperty_field_set_data(self, _get_size_of_list(data), data)

    def as_local_field(self):
        """Create a deep copy of the field locally.

        This copy can then be accessed and modified locally, without a request
        being sent to the server. This method should be used in a ``with``
        statement so that the local field is released and the data is sent to
        the server in one action. If it's not used in a ``with`` statement, the
        method ``release_data()`` should be used to actually update the field.

        .. warning::
           If this ``as_local_field`` method is not used as a context manager in a
           `with` statement or if the method ``release_data()`` is not called,
           the data will not actually be updated.

        Returns
        -------
        local_field : PropertyField

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> num_entities = 5
        >>> field_to_local = dpf.PropertyField(num_entities, dpf.natures.scalar)
        >>> with field_to_local.as_local_field() as f:
        ...     for i in range(1,num_entities+1):
        ...         f.append(list(range(i,i+3)),i)
        ...         print(f.get_entity_data(i-1))
        [1 2 3]
        [2 3 4]
        [3 4 5]
        [4 5 6]
        [5 6 7]


        """
        return _LocalPropertyField(self)


class _LocalPropertyField(_LocalFieldBase, PropertyField):
    """Caches the internal data of a field so that it can be modified locally.

    A single update request is sent to the server when the local field is deleted.

    Parameters
    ----------
    field : PropertyField
        Property field to copy.
    """

    def __init__(self, field):
        self._is_property_field = True
        PropertyField.__init__(self, property_field=field)
        _LocalFieldBase.__init__(self, field)



# -*- coding: utf-8 -*-
"""
MeshesContainer
===============
Contains classes associated with the DPF MeshesContainer.
"""
from ansys.dpf.core import meshed_region
from ansys.dpf.core.collection import Collection
from ansys.dpf.core.plotter import DpfPlotter
from ansys.dpf.core import errors as dpf_errors


class MeshesContainer(Collection):
    """Represents a meshes container, which contains meshes split on a given space.

    Parameters
    ----------
    meshes_container : ansys.grpc.dpf.collection_pb2.Collection or
                       ansys.dpf.core.MeshesContainer, optional
        Create a meshes container from a collection message or create a copy from an
        existing meshes container. The default is ``None``.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    """

    def __init__(self, meshes_container=None, server=None):
        super().__init__(collection=meshes_container, server=server)
        if self._internal_obj is None:
            if self._server.has_client():
                self._internal_obj = self._api.collection_of_mesh_new_on_client(self._server.client)
            else:
                self._internal_obj = self._api.collection_of_mesh_new()

    def create_subtype(self, obj_by_copy):
        return meshed_region.MeshedRegion(mesh=obj_by_copy, server=self._server)

    def plot(self, fields_container=None, deform_by=None, scale_factor=1.0, **kwargs):
        """Plot the meshes container with a specific result if
        fields_container is specified.

        Parameters
        ----------
        fields_container : FieldsContainer, optional
            Data to plot. The default is ``None``.
        deform_by : Field, Result, Operator, optional
            Used to deform the plotted mesh. Must output a 3D vector field.
            Defaults to None.
        scale_factor : float, optional
            Scaling factor to apply when warping the mesh. Defaults to 1.0.
        **kwargs : optional
            Additional keyword arguments for the plotter. For additional keyword
            arguments, see ``help(pyvista.plot)``.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_multishells_rst())
        >>> mesh = model.metadata.meshed_region
        >>> split_mesh_op = dpf.operators.mesh.split_mesh(mesh=mesh, property="mat")
        >>> meshes_cont = split_mesh_op.eval()
        >>> disp_op = dpf.operators.result.displacement(
        ...     data_sources = dpf.DataSources(examples.find_multishells_rst()),
        ...     mesh = meshes_cont
        ... )
        >>> disp_fc = disp_op.outputs.fields_container()
        >>> meshes_cont.plot(disp_fc)

        """
        # DPF defaults
        kwargs.setdefault("show_edges", True)
        # Initiate plotter
        pl = DpfPlotter(**kwargs)
        # If a fields' container is given
        if fields_container is not None:
            for i in range(len(fields_container)):
                label_space = fields_container.get_label_space(i)
                mesh_to_send = self.get_mesh(label_space)
                if mesh_to_send is None:
                    raise dpf_errors.DpfValueError(
                        "Meshes container and result fields "
                        "container do not have the same scope. "
                        "Plotting can not proceed. "
                    )
                field = fields_container[i]
                if deform_by:
                    from ansys.dpf.core.operators import scoping

                    mesh_scoping = scoping.from_mesh(mesh=mesh_to_send)
                    deform_by = deform_by.on_mesh_scoping(mesh_scoping)
                pl.add_field(
                    field,
                    mesh_to_send,
                    deform_by=deform_by,
                    show_axes=kwargs.pop("show_axes", True),
                    scale_factor=scale_factor,
                    **kwargs,
                )
        else:
            # If no field given, associate a random color to each mesh in the container
            from random import random

            random_color = "color" not in kwargs
            for mesh in self:
                if random_color:
                    kwargs["color"] = [random(), random(), random()]
                pl.add_mesh(
                    mesh,
                    deform_by=deform_by,
                    scale_factor=scale_factor,
                    show_axes=kwargs.pop("show_axes", True),
                    **kwargs,
                )
        # Plot the figure
        return pl.show_figure(**kwargs)

    def get_meshes(self, label_space):
        """Retrieve the meshes at a label space.

        Parameters
        ----------
        label_space : dict[str,int]
            Meshes corresponding to a filter (label space) in the input. For example:
            ``{"elshape":1, "body":12}``

        Returns
        -------
        meshes : list[MeshedRegion]
            Meshes corresponding to the request.
        """
        return super()._get_entries(label_space)

    def get_mesh(self, label_space_or_index):
        """Retrieve the mesh at a requested index or label space.

        Raises an exception if the request returns more than one mesh.

        Parameters
        ----------
        label_space_or_index : dict[str,int] , int
            Scoping of the requested mesh, such as ``{"time": 1, "complex": 0}``
            or the index of the mesh.

        Returns
        -------
        mesh : MeshedRegion
            Mesh corresponding to the request.
        """
        return super()._get_entry(label_space_or_index)

    def __getitem__(self, key):
        """Retrieves the mesh at a requested index.

        Parameters
        ----------
        key : int
            Index

        Returns
        -------
        mesh : MeshedRegion
            Mesh corresponding to the request.
        """
        return super().__getitem__(key)

    def add_mesh(self, label_space, mesh):
        """Add or update the scoping at a requested label space.

        Parameters
        ----------
        label_space : dict[str,int]
            Label space of the requested meshes. For example, {"elshape":1, "body":12}.

        mesh : MeshedRegion
            DPF mesh to add or update.
        """
        return super()._add_entry(label_space, mesh)

    def __str__(self):
        txt = "DPF Meshes Container with\n"
        txt += "\t%d mesh(es)\n" % len(self)
        txt += f"\tdefined on labels {self.labels} \n\n"
        return txt



"""
remove_unnecessary_labels
=========================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class remove_unnecessary_labels(Operator):
    """Removes the selected label from a fields container.

    Parameters
    ----------
    permissive : bool, optional
        If true, the operator does not throw an error
        if the label to be removed has more
        than one entry.
    fields_container : FieldsContainer
        Fields container with the label to be
        removed.
    label_to_remove : str
        Label to be removed from the fields
        container.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.remove_unnecessary_labels()

    >>> # Make input connections
    >>> my_permissive = bool()
    >>> op.inputs.permissive.connect(my_permissive)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_label_to_remove = str()
    >>> op.inputs.label_to_remove.connect(my_label_to_remove)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.remove_unnecessary_labels(
    ...     permissive=my_permissive,
    ...     fields_container=my_fields_container,
    ...     label_to_remove=my_label_to_remove,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        permissive=None,
        fields_container=None,
        label_to_remove=None,
        config=None,
        server=None,
    ):
        super().__init__(name="remove_unnecessary_label", config=config, server=server)
        self._inputs = InputsRemoveUnnecessaryLabels(self)
        self._outputs = OutputsRemoveUnnecessaryLabels(self)
        if permissive is not None:
            self.inputs.permissive.connect(permissive)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if label_to_remove is not None:
            self.inputs.label_to_remove.connect(label_to_remove)

    @staticmethod
    def _spec():
        description = """Removes the selected label from a fields container."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                -1: PinSpecification(
                    name="permissive",
                    type_names=["bool"],
                    optional=True,
                    document="""If true, the operator does not throw an error
        if the label to be removed has more
        than one entry.""",
                ),
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fields container with the label to be
        removed.""",
                ),
                1: PinSpecification(
                    name="label_to_remove",
                    type_names=["string"],
                    optional=False,
                    document="""Label to be removed from the fields
        container.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="remove_unnecessary_label", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsRemoveUnnecessaryLabels
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsRemoveUnnecessaryLabels
        """
        return super().outputs


class InputsRemoveUnnecessaryLabels(_Inputs):
    """Intermediate class used to connect user inputs to
    remove_unnecessary_labels operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.remove_unnecessary_labels()
    >>> my_permissive = bool()
    >>> op.inputs.permissive.connect(my_permissive)
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_label_to_remove = str()
    >>> op.inputs.label_to_remove.connect(my_label_to_remove)
    """

    def __init__(self, op: Operator):
        super().__init__(remove_unnecessary_labels._spec().inputs, op)
        self._permissive = Input(
            remove_unnecessary_labels._spec().input_pin(-1), -1, op, -1
        )
        self._inputs.append(self._permissive)
        self._fields_container = Input(
            remove_unnecessary_labels._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._fields_container)
        self._label_to_remove = Input(
            remove_unnecessary_labels._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._label_to_remove)

    @property
    def permissive(self):
        """Allows to connect permissive input to the operator.

        If true, the operator does not throw an error
        if the label to be removed has more
        than one entry.

        Parameters
        ----------
        my_permissive : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remove_unnecessary_labels()
        >>> op.inputs.permissive.connect(my_permissive)
        >>> # or
        >>> op.inputs.permissive(my_permissive)
        """
        return self._permissive

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Fields container with the label to be
        removed.

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remove_unnecessary_labels()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def label_to_remove(self):
        """Allows to connect label_to_remove input to the operator.

        Label to be removed from the fields
        container.

        Parameters
        ----------
        my_label_to_remove : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remove_unnecessary_labels()
        >>> op.inputs.label_to_remove.connect(my_label_to_remove)
        >>> # or
        >>> op.inputs.label_to_remove(my_label_to_remove)
        """
        return self._label_to_remove


class OutputsRemoveUnnecessaryLabels(_Outputs):
    """Intermediate class used to get outputs from
    remove_unnecessary_labels operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.remove_unnecessary_labels()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(remove_unnecessary_labels._spec().outputs, op)
        self._fields_container = Output(
            remove_unnecessary_labels._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.remove_unnecessary_labels()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container
